// -*-Java-*-
//
//  This file is part of the Siena Simulator, a simple discrete-event
//  simulator used with the Siena project.  See
//  http://www.cs.colorado.edu/serl/siena/
//
//  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
//          Matthew J. Rutherford <rutherfo@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 1998-2003 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: Process.java,v 1.2 2003/04/22 21:26:44 mjr Exp $
//
package ssim;

/**  Virtual class (interface) representing processes running within
 *  the simulator.
 *
 *  A simulated process must implement this basic interface.
 **/
public interface Process
{
    public final static long NULL_PROCESSID = 0;

    /**  action when the process is created.
     *
     *  This execution step is scheduled when the process is created
     *  through Sim::create_process.  This action is executed
     *  before any event or timeout.
     **/
    public void init();

    /** action executed in response to an event signaled to this process.
     *
     *  Notice that the signaled event should not be used outside this
     *  method, other than by signaling it to a process through
     *  Sim::signal_event.  Specifically, the Event pointer passed to
     *  this method should be considered invalid outside this method.
     *  in fact, the simulator proceeds to de-allocate every event
     *  object that is not signaled to any other process.
     *
     *  <p>The implementation of this method may specify the duration
     *  of the actions associated with this response using the \link
     *  Sim::advance_delay(Time) advance_delay method\endlink.  By
     *  default, the duration of an action is 0.
     *
     *  @see Sim#signal_event(Event,long)
     *  @see Sim#signal_event(Event,long,long)
     *  @see Sim#advance_delay(long)
     **/
    void process_event( Event evt );

    /** action executed in response to a timeout.
     *
     *  This method defines the actions explicitly scheduled for this
     *  process by the process itself.  These actions are scheduled by
     *  calling Sim::set_timeout(Time).
     *
     *  <p>The implementation of this method may specify the duration
     *  of the actions associated with this response using the \link
     *  Sim::advance_delay(Time) advance_delay method\endlink.  By
     *  default, the duration of an action is 0.
     **/
    void process_timeout();

    /**  executed when the process is explicitly stopped.
     *
     *  A process is stopped by a call to
     *  Sim::stop_process(ProcessId).  This method is executed
     *  immediately after the process has processed all the events or
     *  timeouts scheduled before the call to Sim::stop_process.
     **/
    void stop();
}
