// -*-Java-*-
//
//  This file is part of the Siena Simulator, a simple discrete-event
//  simulator used with the Siena project.  See
//  http://www.cs.colorado.edu/serl/siena/
//
//  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
//          Matthew J. Rutherford <rutherfo@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 1998-2003 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: Sim.java,v 1.2 2003/04/22 21:26:44 mjr Exp $
//
package ssim;

import java.util.HashMap;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Vector;


/**  a generic discrete-event sequential simulator
 *
 *  This class implements a generic discrete-event sequential
 *  simulator.  Sim maintains and executes a time-ordered schedule of
 *  actions (or discrete events).
 **/ 
public class Sim
{
    /**  <code>P_SEQUENTIAL</code> flag */
    public final static char P_SEQUENTIAL = 0x02;	
    /**  <code>P_QUEUEING</code> flag */
    public final static char P_QUEUEING = 0x04; 
    /** Initial time. */
    public final static long INIT_TIME = 0;

    /**  virtual time 
     *
     *  time in the simulated world.
     **/
    private static long m_current_time = INIT_TIME;
    /** currently running process **/ 
    private static long m_current_process = Process.NULL_PROCESSID;
    /** duration of the current action **/
    private static long	m_current_delay = INIT_TIME;
    /** is simulation running? **/
    private static boolean m_running = false;

    private final static short A_Event = 0;
    private final static short A_Timeout = 1;
    private final static short A_Init = 2;
    private final static short A_Stop = 3;

    static class Action
    {
	public short type;
	long pid;
	Event event = null;
	
	public Action( short p_type, long p_pid )
	{
	    type = p_type;
	    pid = p_pid;
	}

	public Action( short p_type, long p_pid, Event p_event )
	{
	    type = p_type;
	    pid = p_pid;
	    event = p_event;
	}

	public Action( Action a )
	{
	    type = a.type;
	    pid = a.pid;
	    event = a.event;
	}

	public void assign( Action a )
	{
	    type = a.type;
	    pid = a.pid;
	    event = a.event;
	}
    }

//     typedef multimap<Time, Action>	ActionsTable;
//     static ActionsTable			actions;
    private static SortedMap m_actions = new TreeMap();

    private static void	schedule( Action a, long t )
    {
	Long key = new Long( clock() + t );
	Vector acts;
	if( !m_actions.containsKey( key ) ) {
	    acts = new Vector();
	    m_actions.put( key, acts );
	}
	else {
	    acts = (Vector)m_actions.get( key );
	}
	acts.addElement( a );
    }

    private static void	schedule_now( Action a )
    {
	schedule( a, m_current_time );
    }

    //
    // internal process status (within Sim)
    // 
    private static final short P_TERMINATED	= 0x01;	// running(0)/terminated(1) 

    static class PDescr
    {
	public Process process = null;
	public short	status = P_TERMINATED;
	public long total_action_time = INIT_TIME;
	public long available_at = INIT_TIME;

	public PDescr()
	{
	}

	public PDescr(Process p, short mode )
	{
	    process = p;
	    status = mode;
	}
    }

//     typedef vector<PDescr>		PsTable;
//     static PsTable			processes;
    private static HashMap m_processes = new HashMap();

    /**  creates a new process
     *
     *  creates a new process with the given Process object.
     *  <code>mode</code> specifies the execution flags of the
     *  process.  If the <code>P_SEQUENTIAL</code> flag is set, then
     *  the process will process one event at a time with respect to
     *  the simulation virtual time, otherwise (default) the actions
     *  of this process will be considered reentrant and may be
     *  executed in parallel.  By default, the simulator discards the
     *  events signalled to a <code>P_SEQUENTIAL</code> process while
     *  that process is busy executing other actions.  The
     *  <code>P_QUEUEING</code> flag can be set to instruct the
     *  simulator to queue those events.  In this case, the simulator
     *  will deliver signals to that process as the process is
     *  available to respond to them.
     **/
    public static long create_process( Process proc )
    {
	return create_process( proc, (short)0 );
    }

    public static long create_process( Process proc, short mode )
    {
	mode &= ( P_SEQUENTIAL | P_QUEUEING );
	long newpid = m_processes.size();
	m_processes.put( new Long( newpid ), new PDescr( proc, mode ) );
	schedule_now( new Action( A_Init, newpid ) );
	return newpid;
    }

    /**  stops the execution of a given process */
    public static int stop_process( long pid )
    {
	PDescr pd = (PDescr)m_processes.get( new Long( pid ) );
	if( (pd.status & P_TERMINATED) > 0 ) {
	    return -1;
	}
	schedule_now( new Action( A_Stop, pid ) );
	return 0;
    }

    /**  stops the execution of the current process */
    public static void stop_process()
    {
	schedule_now( new Action( A_Stop, m_current_process ) );
    }

   /**  clears out internal data structures
    *
    *  Resets the simulator making it available for a completely new
    *  simulation.  All scheduled actions are deleted together with
    *  the associated events.  All process identifiers returned by
    *  previoius invocations of \link create_process(Process*,char)
    *  create_process\endlink are invalidated by this operation.
    *  Notice however that it is the responsibility of the simulation
    *  programmer to delete process objects used in the simulation.
    **/
    public static void clear()
    {
	m_running = false;
	m_current_time = INIT_TIME;
	m_current_delay = INIT_TIME;
	m_processes = new HashMap();
	m_actions = new TreeMap();
    }

    /**  signal an event to the given process 
     *
     *  Signal an event to the given process.  The response is
     *  scheduled for the current time.
     *
     *  @see Process#process_event(Event) Process.process_event(Event)
     **/
    public static int signal_event( Event evt, long pid )
    {
	PDescr pd = (PDescr)m_processes.get( new Long( pid ) );
	if( (pd.status & P_TERMINATED) > 0 ) {
	    return -1;
	}
	schedule_now( new Action( A_Event, pid, evt ) );
	return 0;
    }

    /**  signal an event to the given process at the given time 
     *
     *  @see Process#process_event(Event) process_event
     **/
    public static int signal_event( Event evt, long pid, long delay )
    {
	PDescr pd = (PDescr)m_processes.get( new Long( pid ) );
	if( (pd.status & P_TERMINATED) > 0 ) {
	    return -1;
	}
	schedule( new Action( A_Event, pid, evt ), delay );
	return 0;
    }

    /**  sets a timeout for the current process
     *
     *  Schedules the execution of process_timeout() on the current
     *  process after the given amount of (virtual) time.
     **/
    public static void set_timeout( long t )
    {
	schedule( new Action( A_Timeout, m_current_process ), t );
    }

    /**  Advance the execution time of the current process.
     *
     *  This method can be used to specify the duration of certain
     *  actions, or certain steps within the same action.  For example:
     *
     *  <pre>
     *  class SomeProcess : public Process {
     *      //...
     *      virtual void process_event(const Event * e) {
     *          //
     *          // suppose this response is called at (virtual) time X
     *
     *          // ...do something here...
     *
     *          // the above actions have a default duration of 0,
     *          // therefore the following event is scheduled at time X + 5
     *          //
     *          signal_event(e, p, 5);
     *     
     *          advance_delay(10);
     *
     *          // ...do something else here...
     *
     *          // advance_delay(10) specifies a duration of 10, therefore 
     *          // the following timeout is scheduled at time X + 15;
     *          //
     *          set_timeout(5);
     *      }
     *  };
     *  </pre>
     *
     *  @see Sim#clock() Sim.clock()
     **/
    public static void advance_delay( long delay )
    {
	if( !m_running ) {
	    return;
	}
	PDescr pd = (PDescr)m_processes.get( new Long( m_current_process ) );
	pd.total_action_time += delay;
	m_current_delay +=  delay;
    }
    
    /**  Returns the current process.
     *
     *  Example:
     *  <pre>
     *  void SimpleProcess::process_event(const Event *e) {
     *      cout << "my process id is: " << Sim::this_pocess() << endl;
     *  }
     *  </pre>
     **/
    public static long this_process()
    {
	return m_current_process;
    }

    /**  Returns the current virtual time for the current process.
     *  
     *  Example:
     *  <pre>
     *  void LoopProcess::process_timeout() {
     *      cout << "Here I am at time:" << Sim::clock() << endl;
     *      set_timeout(20)
     *      cout << "I just scheduled myself again for time: " 
     *           << Sim::clock() + 20 << endl;
     *  }
     *  </pre>
     *
     *  @see Sim#advance_delay(long) Sim.advance_delay(long)
     **/
    public static long clock()
    {
	return m_current_time + m_current_delay;
    }
    
    private static boolean m_lock = false;

    /**  starts execution of the simulation */
    public static void run_simulation()
    {
	//
	// prevents anyone from re-entering the main loop.  Note that this
	// isn't meant to be thread-safe, it works if some process calls
	// Sim::run_simulation() within their process_event() or
	// process_timeout() function.
	//
	if( m_lock ) {
	    return;
	}
	m_lock = true;
	m_running = true;

	//
	// while there is at least a scheduled action
	//
	while( m_running && !m_actions.isEmpty() ) {
	    //
	    // I'm purposely excluding any kind of checks in this version
	    // of the simulator.  
	    //
	    // I should say something like this:
	    // assert(current_time <= (*a).first);
	    //
	    Long key = (Long)m_actions.firstKey();
	    m_current_time = key.longValue();
	    Vector actions = (Vector)m_actions.get( key );
	    Action action = (Action)actions.firstElement();
	    actions.removeElementAt( 0 );
	    if( actions.isEmpty() ) {
		m_actions.remove( key );
	    }
	    m_current_process = action.pid;
	    m_current_delay = 0;
	    //
	    // right now I don't check if current_process is indeed a
	    // valid process.  Keep in mind that this is the heart of the
	    // simulator main loop, therefore efficiency is crucial.
	    // Perhaps I should check.  This is somehow a design choice.
	    //
	    PDescr pd = (PDescr)m_processes.get( new Long( m_current_process ) );
	    
	    if( (pd.status & P_TERMINATED) > 0 ) {
		//
		// ...work in progress...
		//
		action.event = null;
	    }
	    else if( (pd.status & P_SEQUENTIAL) > 0 && 
		     m_current_time < pd.available_at ) {
		//
		// this process is sequential and is currently executing
		// another action (in virtual time of course), so we
		// reschedule this action to the time the process is
		// available
		//
		if( (pd.status & P_QUEUEING) > 0 ) {
		    schedule( action, pd.available_at );
		}
		//
		// if the process can not queue actions, this action is
		// simply dropped
		//
	    }
	    else {
		switch( action.type ) {
		case A_Event:
		    pd.process.process_event( action.event );
		    break;
		case A_Timeout:
		    pd.process.process_timeout();
		    break;
		case A_Init:
		    pd.process.init();
		    break;
		case A_Stop: 
		    pd.process.stop();
		    pd.status |= P_TERMINATED;
		    break;
		default:
		    //
		    // add paranoia checks/logging here?
		    //
		    break;
		}
		if( (pd.status & P_SEQUENTIAL) > 0 ) {
		    pd.available_at = clock();
		}
	    }
	}
	m_lock = false;
    }
    
    /**  stops execution of the simulation */
    public static void stop_simulation()
    {
	m_running = false;
    }
}
